import math
from typing import Generic, Dict, Any, Type, Sequence, Optional, cast
from uuid import uuid4
import time

from jwt import decode as jwt_decode, encode as jwt_encode, PyJWTError


from jotbox.types import EncodedToken, StrBytes, TPayload, Payload
from jotbox.json import make_json_encoder
from jotbox.whitelist.base import Whitelist
from jotbox.exceptions import JWTDecodeError, RevokedTokenError


class Jotbox(Generic[TPayload]):
    whitelist: Optional[Whitelist]

    def __init__(
        self,
        *,
        encode_key: StrBytes,
        payload_type: Type[TPayload] = cast(Type[TPayload], Payload),
        decode_key: Optional[StrBytes] = None,
        encode_algorithm: str = "HS512",
        decode_algorithms: Sequence[str] = ("HS512",),
        leeway: int = 0,
        expires_in: Optional[int] = None,
        idle_timeout: Optional[int] = None,
        whitelist: Optional[Whitelist[TPayload]] = None,
        jwt_options: Dict[str, bool] = None,
    ) -> None:
        if idle_timeout is not None and whitelist is None:
            raise ValueError("idle_timeout is not possible without whitelist")
        self.payload_type = payload_type
        self.encode_key = encode_key
        self.decode_key = decode_key if decode_key is not None else encode_key
        self.encode_algorithm = encode_algorithm
        self.decode_algorithms = decode_algorithms
        self.leeway = leeway
        self.expires_in = expires_in
        self.idle_timeout = idle_timeout
        self.jwt_options = jwt_options
        self.whitelist = whitelist

        self.encoded_token_type = EncodedToken[self.payload_type]  # type: ignore

    def _create_payload_dict(self, **claims) -> Dict[str, Any]:
        iat = math.floor(time.time())
        claims = dict(claims, jti=uuid4(), iat=iat)
        if self.expires_in is not None and "exp" not in claims:
            claims["exp"] = claims["iat"] + self.expires_in
        return claims

    def create_payload(self, **claims) -> TPayload:
        return self.payload_type(**self._create_payload_dict(**claims))

    def _create_until(self, exp: Optional[int]) -> Optional[float]:
        if self.idle_timeout is None:
            return exp + self.leeway if exp is not None else None
        else:
            idle_until = time.time() + self.idle_timeout + self.leeway
        if exp is None:
            return idle_until
        return min(idle_until, exp + self.leeway)

    async def add_to_whitelist(self, payload: TPayload) -> None:
        if not self.whitelist:
            return
        until = self._create_until(payload.exp)
        await self.whitelist.add(payload, until)

    async def create_token(
        self, _payload: Optional[TPayload] = None, **claims
    ) -> EncodedToken[TPayload]:
        payload = _payload or self.create_payload(**claims)
        token = jwt_encode(
            payload.dict(exclude_unset=True),
            key=self.encode_key,
            algorithm=self.encode_algorithm,
            json_encoder=make_json_encoder(self.payload_type),
        ).decode()
        await self.add_to_whitelist(payload)
        return self.encoded_token_type(token=token, payload=payload)

    async def verified_payload(self, token: str, **jwt_kwargs) -> TPayload:
        try:
            raw_payload = jwt_decode(
                token,
                key=self.decode_key,
                algorithms=self.decode_algorithms,
                options=self.jwt_options,
                leeway=self.leeway,
                **jwt_kwargs,
            )
        except PyJWTError as e:
            raise JWTDecodeError("Failed to decode token") from e
        payload = self.payload_type(**raw_payload)
        await self._verify_whitelist(payload)
        return payload

    async def _verify_whitelist(self, payload: TPayload) -> None:
        if not self.whitelist:
            return
        if self.idle_timeout:
            until = cast(float, self._create_until(payload.exp))
            exists = await self.whitelist.touch(payload, until)
        else:
            exists = await self.whitelist.exists(payload)
        if not exists:
            raise RevokedTokenError(f"Token ID {payload.jti} has been revoked")

    async def revoke_payload(self, payload: TPayload) -> None:
        if not self.whitelist:
            raise NotImplementedError(
                "This operation is not supported without whitelist"
            )
        await self.whitelist.delete(payload)



from jotbox.tokens import Jotbox
from jotbox.types import (
    TPayload,
    TExpiringPayload,
    ExpiringPayload,
    Payload,
    EncodedToken,
    TSub,
    TSession,
    Session,
)
from jotbox.sessions import SessionBox
from jotbox.exceptions import (
    JotboxError,
    JWTDecodeError,
    RevokedTokenError,
)



from typing import Union, Optional

from aredis import StrictRedis


from jotbox.types import TPayload, DateTimeStamp, TSession, TSub
from jotbox.util import timestamp_ms, now_ms
from jotbox.whitelist.base import Whitelist, SessionWhitelist
from jotbox.util import achunked


RedisOrURL = Union[StrictRedis, str]


class RedisWhitelist(Whitelist[TPayload]):
    def __init__(self, redis: RedisOrURL, key_prefix: str = "JB:WHITE:") -> None:
        self.key_prefix = key_prefix
        if isinstance(redis, str):
            self.redis = StrictRedis.from_url(redis)
        else:
            self.redis = redis

    async def add(self, payload: TPayload, until: Optional[DateTimeStamp]) -> None:
        px = None
        if until:
            px = timestamp_ms(until) - now_ms()
        key = self.make_key(payload)
        await self.redis.set(key, 1, px=px, nx=True)

    async def touch(self, payload: TPayload, until: DateTimeStamp) -> bool:
        key = self.make_key(payload)
        px = timestamp_ms(until) - now_ms()
        return bool(await self.redis.pexpire(key, px))

    async def exists(self, payload: TPayload) -> bool:
        return bool(await self.redis.exists(self.make_key(payload)))

    async def delete(self, payload: TPayload) -> None:
        await self.redis.delete(self.make_key(payload))

    def make_key(self, payload: TPayload) -> str:
        return self.key_prefix + str(payload.jti)


class SessionRedisWhitelist(RedisWhitelist[TSession], SessionWhitelist[TSession, TSub]):
    def make_key(self, payload: TSession) -> str:
        return self.make_sub_key(payload.sub) + str(payload.jti)

    def make_sub_key(self, sub: TSub) -> str:
        return self.key_prefix + f"{sub}:"

    async def delete_sub(self, sub: TSub) -> None:
        async for keys in achunked(
            self.redis.scan_iter(self.make_sub_key(sub) + "*"), 500
        ):
            await self.redis.delete(*keys)



from typing import TypeVar, Generic, Type, cast

from pydantic.generics import GenericModel

from jotbox.tokens import Jotbox
from jotbox.types import TSession, TSub, Session
from jotbox.whitelist.base import SessionWhitelist


class SessionBox(Jotbox[TSession], Generic[TSession, TSub]):
    whitelist: SessionWhitelist[TSession, TSub]

    def __init__(
        self,
        *,
        whitelist: SessionWhitelist,
        payload_type: Type[TSession] = cast(Type[TSession], Session),
        **kwargs
    ) -> None:
        super().__init__(whitelist=whitelist, payload_type=payload_type, **kwargs)
        if not self.whitelist:
            raise ValueError("whitelist argument is required")

    async def revoke_subject(self, sub: TSub) -> None:
        """
        Revoke all tokens with given `sub` claim
        """
        await self.whitelist.delete_sub(sub)



from datetime import datetime
from typing import TypeVar, Generic, Union, Optional
from uuid import UUID

from pydantic import BaseModel
from pydantic.generics import GenericModel


StrBytes = Union[str, bytes]


class Payload(BaseModel):
    class Config:
        json_encoders = {UUID: lambda v: v.hex}

    jti: UUID
    iat: int
    exp: Optional[int] = None


class ExpiringPayload(Payload):
    exp: int


TPayload = TypeVar("TPayload", bound=Payload)
TExpiringPayload = TypeVar("TExpiringPayload", bound=ExpiringPayload)


class EncodedToken(GenericModel, Generic[TPayload]):
    payload: TPayload
    token: str


DateTimeStamp = Union[datetime, float]


TSub = TypeVar("TSub")


class Session(GenericModel, Generic[TSub], Payload):
    sub: TSub


TSession = TypeVar("TSession", bound=Session)



from math import floor
from datetime import datetime
import time
from typing import AsyncIterable, TypeVar, List, AsyncIterator

from jotbox.types import DateTimeStamp


T = TypeVar("T")


def timestamp_ms(t: DateTimeStamp) -> int:
    if isinstance(t, datetime):
        return floor(t.timestamp() * 1000)
    return floor(t * 1000)


def now_ms() -> int:
    return timestamp_ms(time.time())


async def achunked(iterable: AsyncIterable[T], size: int) -> AsyncIterator[List[T]]:
    chunk = []
    async for item in iterable:
        chunk.append(item)
        if len(chunk) == size:
            yield chunk
            chunk = []
    if chunk:
        yield chunk



import json
from typing import Type

from jotbox.types import TPayload


_JSON_ENCODERS: dict = {}


def make_json_encoder(t: Type[TPayload]):
    if t not in _JSON_ENCODERS:
        _JSON_ENCODERS[t] = type(
            "CustomJsonEncoder",
            (json.JSONEncoder,),
            {"default": lambda self, x: t.__json_encoder__(x)},
        )
    return _JSON_ENCODERS[t]



from abc import abstractmethod, ABC
from typing import Optional, Generic

from jotbox.types import TPayload, DateTimeStamp, TSub, TSession


class Whitelist(ABC, Generic[TPayload]):
    @abstractmethod
    async def add(self, payload: TPayload, until: Optional[DateTimeStamp]) -> None:
        """
        Save a token payload in the whitelist.

        The `until` timestamp if not `None` must be respected
        and the token should be considered expired at this time.

        If `until` is `None` the token should not expire
        """

    @abstractmethod
    async def exists(self, payload: TPayload) -> bool:
        """
        Check whether the token exists and is currently valid in the whitelist.

        return `True` if the token is valid and not expired, otherwise `False`
        """

    @abstractmethod
    async def touch(self, payload: TPayload, until: DateTimeStamp) -> bool:
        """
        Used for the "idle timeout" function.

        Revalidate the token and extend its expiry time up to `until` timestamp
        This method must verify that the token is currently valid before
        revalidation.

        return
           `True` if the token is succesfully "touched"
           `False` if the token is not currently valid in the whitelist.
        """

    @abstractmethod
    async def delete(self, delete: TPayload) -> None:
        """
        Immediately revoke the token from the whitelist
        """


class SessionWhitelist(Whitelist[TSession], Generic[TSession, TSub]):
    @abstractmethod
    async def delete_sub(self, sub: TSub) -> None:
        """
        Delete all tokens with a given `sub` claim
        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jotbox', 'jotbox.sessions', 'jotbox.whitelist']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1,<2', 'pyjwt>=1,<2']

setup_kwargs = {
    'name': 'jotbox',
    'version': '0.1.1',
    'description': 'JWT library with support for revokable tokens',
    'long_description': '# Jotbox\n\nJWT library for python with support for revokable tokens and idle timeouts\n\nRead more here https://steinitzu.github.io/jotbox\n',
    'author': 'Steinthor Palsson',
    'author_email': 'steini90@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
